<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8">
  <title>{{DP3_APP}}</title>
  <style>
    body {
      text-align: center;
      margin: 0;
    }
    .centered-div {
      width: 50%;
      margin: 0 auto;
      text-align: left;
    }
  </style>
</head>
<body>

  <div class="centered-div">
    <h1>An {{DP3_APP}} server</h1>

    <p><a href="http://{{HOSTNAME}}/api/">DP3 API</a></p>
    <p><a href="http://{{HOSTNAME}}/api/docs">DP3 API docs</a></p>
    <p><a href="http://{{HOSTNAME}}/rabbitmq/">RabbitMQ management interface</a></p>
  </div>

</body>
</html>