"""
Common modules which are used throughout the platform.

- [Config][dp3.common.config], [EntitySpec][dp3.common.entityspec] and
[AttrSpec][dp3.common.attrspec] - Models for reading, validation and representing
platform configuration of entities and their attributes.
[datatype][dp3.common.datatype] is also used within this context.
- [Scheduler][dp3.common.scheduler] - Allows modules to run callbacks at specified times
- [Task][dp3.common.task.Task] - Model for a single task processed by the platform
- [Utils][dp3.common.utils] - Auxiliary utility functions
"""
