"""
[SnapShooter][dp3.snapshots.snapshooter], a module responsible
for snapshot creation and running configured data correlation and fusion hooks,
and [Snapshot Hooks][dp3.snapshots.snapshot_hooks], which manage the registered hooks and their
dependencies on one another.
"""
